from datetime import date, timedelta
from typing import List, Tuple, cast

import pandas as pd

from fao56_model.crop_calendar import CropCalendar
from fao56_model.data_models.climate import ClimateData
from fao56_model.data_models.simulation_state import SimulationState
from fao56_model.data_models.soil import Soil


class CropSimulation:
    def __init__(
        self, calendar: CropCalendar, soil: Soil, climate_date: ClimateData
    ):
        self.calendar = calendar
        self.climate_data = climate_date
        self.soil = soil
        if not self.climate_data_is_within_range():
            raise ValueError(
                "Climate data does not cover the crop calendar range"
            )
        self.initialize_states()

    def climate_data_is_within_range(self) -> bool:
        """
        Check if the data is within the crop calendar range
        """
        return self.calendar.time_range in self.climate_data.time_range

    def get_total_available_water(self, rooting_depth: float) -> float:
        """
        Returns the total available water for the given rooting depth
        """
        return (
            1000
            * (self.soil.theta_fc - self.soil.theta_wp)
            * 0.5
            * rooting_depth
        )

    def initialize_states(self) -> None:
        """
        Initialize the states with the climate data and the soil data
        """
        self.states = [
            SimulationState(
                timestamp=d.timestamp,
                ET_0=d.ET_0,
                precipation_mm=d.precipation_mm,
            )
            for d in self.climate_data.data_points
        ]
        # initialize the root zone depletion to
        # the readily available water one day before the sowing date
        crop_start_times = [c.sowing_date for c in self.calendar.planted_crops]
        for i, state in enumerate(self.states):
            if state.timestamp in crop_start_times:
                crop = self.calendar.get_crop(state.timestamp)
                assert crop
                self.states[i - 1].Z_r = crop.crop.min_root_depth_m
                self.states[i - 1].D_r = self.get_total_available_water(
                    crop.crop.min_root_depth_m
                )
                break

    def run(self) -> None:
        """
        Run the simulation
        """
        for i, state in enumerate(self.states[1:]):
            idx = i + 1
            prev_state = self.states[idx - 1]
            planted_crop = self.calendar.get_crop(state.timestamp)
            if planted_crop:
                assert prev_state.D_r is not None
                state.ET_c = (
                    planted_crop.get_crop_coefficient(state.timestamp)
                    * state.ET_0
                )
                state.d = planted_crop.crop.const_depletion_factor + 0.04 * (
                    5 - state.ET_c
                )
                state.Z_r = planted_crop.get_rooting_depth(state.timestamp)
                state.TAW = self.get_total_available_water(state.Z_r)
                state.RAW = state.d * state.TAW
                state.D_r = (
                    state.TAW
                    * (
                        -prev_state.D_r * state.d
                        + prev_state.D_r
                        + state.ET_c
                        + state.I_del * state.d
                        - state.I_del
                        + state.precipation_mm * state.d
                        - state.precipation_mm
                        - state.RO * state.d
                        + state.RO
                    )
                ) / (state.ET_c - state.TAW * state.d + state.TAW)
                state.D_r = max(min(state.D_r, state.TAW), 0)
                state.K_s = 1
                if state.D_r > state.RAW:
                    state.K_s = (state.TAW - state.D_r) / (
                        (1 - state.d) * state.TAW
                    )
                state.ET_a = state.K_s * state.ET_c
                state.ET_rel = state.ET_a / state.ET_0
                state.ky = 1

    def calc_yield(self) -> List[Tuple[date, float]]:
        """
        Calculate the yield
        """
        crops = [crop for crop in self.calendar.planted_crops]
        harvest_dates = [c.harvest_date for c in self.calendar.planted_crops]
        ls_actual_yields = []
        for crop in crops:
            period = crop.growing_period
            timestamps = [d for d in period.range(timedelta(days=1))]
            ky_values = [
                cast(float, s.ky)
                for s in self.states
                if s.timestamp in timestamps
            ]
            avg_ky = sum(ky_values) / len(ky_values)
            actual_yield = crop.crop.yield_max * avg_ky
            ls_actual_yields.append(actual_yield)
        return list(zip(harvest_dates, ls_actual_yields))

    @property
    def df(self) -> pd.DataFrame:
        return pd.DataFrame([s.model_dump() for s in self.states])



from typing import Literal

from pydantic import BaseModel, Field

SoilType = Literal[
    "sand",
    "loamy sand",
    "sandy loam",
    "loam",
    "silt loam",
    "silt",
    "silt clay loam",
    "silty clay",
    "clay",
]


class Soil(BaseModel):
    kind: SoilType = Field(description="Soil type", alias="soil_type")
    theta_fc_max: float = Field(
        description="Water content at field capacity (maximum)"
    )
    theta_fc_min: float = Field(
        description="Water content at field capacity (minimum)"
    )
    theta_wp_max: float = Field(
        description="Water content at wilting point (maximum)"
    )
    theta_wp_min: float = Field(
        description="Water content at wilting point (minimum)"
    )

    @property
    def theta_fc(self):
        """
        Water content at field capacity (average)
        """
        return (self.theta_fc_max + self.theta_fc_min) / 2

    @property
    def theta_wp(self):
        """
        Water content at wilting point (average)
        """
        return (self.theta_wp_max + self.theta_wp_min) / 2



from typing import List

from datetimerange import DateTimeRange


def is_overlapping(list_range: List[DateTimeRange]) -> bool:
    """
    Returns True if list of DateTimeRange objects has any overlapping ranges
    """
    for i in range(len(list_range)):
        for j in range(i + 1, len(list_range)):
            if list_range[i].is_intersection(list_range[j]):
                return True
    return False



from datetime import datetime
from typing import Optional

from pydantic import BaseModel, Field


class SimulationState(BaseModel):
    timestamp: datetime = Field(..., description="Timestamp of the simulation")
    ET_0: float = Field(
        100, description="Reference evapotranspiration [mm/day]"
    )
    ET_c: Optional[float] = Field(
        None, description="Crop evapotranspiration [mm/day]"
    )
    d: Optional[float] = Field(None, description="Depletion fraction [-]")
    TAW: Optional[float] = Field(
        None, description="Total available water [mm]"
    )
    RO: float = Field(0, description="Runoff [mm/day]")
    RAW: Optional[float] = Field(
        None, description="Readily available water [mm]"
    )
    I_dem: Optional[float] = Field(
        None, description="Irrigation demand [mm/day]"
    )
    I_del: float = Field(0, description="Irrigation demand [mm/day]")
    K_s: Optional[float] = Field(
        None, description="Soil water stress coefficient [-]"
    )
    ET_a: Optional[float] = Field(
        None, description="Actual evapotranspiration [mm/day]"
    )
    ky: Optional[float] = Field(None, description="Yield response factor [-]")
    Z_r: Optional[float] = Field(None, description="Root zone depth [mm]")
    D_r: Optional[float] = Field(None, description="Root zone depletion [mm]")
    ET_rel: Optional[float] = Field(
        None, description="Relative evapotranspiration [-]"
    )
    precipation_mm: float = Field(0, description="Precipitation [mm]")



from datetime import date, datetime, timedelta
from typing import Dict, Literal, cast

from datetimerange import DateTimeRange

from fao56_model.data_models.crop import Crop

CropStage = Literal["ini", "dev", "mid", "end"]


class PlantedCrop(object):
    def __init__(self, crop: Crop, sowing_date: date):
        self.crop = crop
        self._set_dates(datetime.combine(sowing_date, datetime.min.time()))

    def _set_dates(self, sowing_date: datetime) -> None:
        self.sowing_date = sowing_date
        initial_end = self.sowing_date + timedelta(
            days=self.crop.growth_stage_ini_d
        )
        development_end = initial_end + timedelta(
            days=self.crop.growth_stage_dev_d
        )
        mid_end = development_end + timedelta(
            days=self.crop.growth_stage_mid_d
        )
        end_end = mid_end + timedelta(days=self.crop.growth_stage_end_d)

        self.initial_stage = DateTimeRange(self.sowing_date, initial_end)
        self.development_stage = DateTimeRange(initial_end, development_end)
        self.mid_stage = DateTimeRange(development_end, mid_end)
        self.end_stage = DateTimeRange(mid_end, end_end)
        self.growing_period = (
            self.initial_stage.encompass(self.development_stage)
            .encompass(self.mid_stage)
            .encompass(self.end_stage)
        )
        self.harvest_date = cast(
            datetime, self.growing_period.end_datetime
        ).date()
        self.stages: Dict[CropStage, DateTimeRange] = {
            "ini": self.initial_stage,
            "dev": self.development_stage,
            "mid": self.mid_stage,
            "end": self.end_stage,
        }

    def get_crop_coefficient(self, current_date: date) -> float:
        """
        Returns the crop coefficient for the given date
        """
        current_datetime = datetime.combine(current_date, datetime.min.time())
        stage = self.get_crop_stage(current_datetime)
        coefficients: Dict[CropStage, float] = {
            "ini": self.crop.crop_coefficient_ini,
            "mid": self.crop.crop_coefficient_mid,
            "end": self.crop.crop_coefficient_end,
        }
        if stage == "dev":
            current_timedelta = DateTimeRange(
                self.initial_stage.end_datetime, current_datetime
            ).timedelta
            dev_timedelta = self.development_stage.timedelta
            return self.crop.crop_coefficient_ini + (
                current_timedelta / dev_timedelta
            ) * (
                self.crop.crop_coefficient_mid - self.crop.crop_coefficient_ini
            )
        else:
            return coefficients[stage]

    def get_crop_stage(self, current_datetime: datetime) -> CropStage:
        """
        Determine the crop stage based on the current date.
        """
        stages: Dict[CropStage, DateTimeRange] = {
            "ini": self.initial_stage,
            "dev": self.development_stage,
            "mid": self.mid_stage,
            "end": self.end_stage,
        }
        for stage_name, stage_range in stages.items():
            if current_datetime in stage_range:
                return stage_name
        raise ValueError("Date is not within the growing period")

    def get_corrected_crop_coefficient_w_correlation(
        self, default_crop_coefficient: float, u_mean_ms: float, RH_min: float
    ) -> float:
        """
        Crop coefficient correlation
        """
        return default_crop_coefficient + (
            0.04 * (u_mean_ms - 2) - 0.004 * (RH_min - 45)
        ) * (self.crop.max_crop_height_m / 3)

    def get_rooting_depth(self, current_date: datetime) -> float:
        """
        Returns the rooting depth [meters]
        """
        if current_date not in self.growing_period:
            raise ValueError("Date is not within the growing period")
        t = (current_date - self.sowing_date).days
        t_x = self.crop.sowing_to_max_rooting_depth_d
        t_0 = self.crop.sowing_to_emergence_d
        z_0 = (
            self.crop.start_root_depth_pct / 100.0 * self.crop.min_root_depth_m
        )

        z_max = self.crop.max_root_depth_m
        z_min = self.crop.min_root_depth_m
        z = z_0 + (z_max - z_0) * (t - t_0 / 2) / (t_x - t_0 / 2)
        z_r = min(max(z_min, z), z_max)
        return z_r



# -*- coding: utf-8 -*-

"""Package for fao56_model"""

__author__ = "Stephan Smit"
__email__ = "stephansmit@hotmail.com"
# Do not edit this string manually, always use bumpversion
# Details in CONTRIBUTING.md
__version__ = "0.1.1"


def get_module_version():
    return __version__



from typing import Literal

from pydantic import BaseModel, Field

CropType = Literal["spring wheat"]


class Crop(BaseModel):
    kind: CropType = Field(..., description="Crop type", alias="crop_type")
    crop_coefficient_ini: float = Field(
        ..., description="Initial crop coefficient", alias="kc_ini"
    )
    crop_coefficient_mid: float = Field(
        ..., description="Midseason crop coefficient", alias="kc_mid"
    )
    crop_coefficient_end_min: float = Field(
        ...,
        description="Endseason crop coefficient (minimum)",
        alias="kc_end_min",
    )
    crop_coefficient_end_max: float = Field(
        ...,
        description="Endseason crop coefficient (maximum)",
        alias="kc_end_max",
    )
    growth_stage_ini_d: int = Field(
        ..., description="Length of initial growth stage [days]", alias="l_ini"
    )
    growth_stage_dev_d: int = Field(
        ...,
        description="Length of development growth stage [days]",
        alias="l_dev",
    )
    growth_stage_mid_d: int = Field(
        ..., description="Length of middle growth stage [days]", alias="l_mid"
    )
    growth_stage_end_d: int = Field(
        ..., description="Length of ending growth stage [days]", alias="l_end"
    )
    max_crop_height_m: float = Field(
        ..., description="Maximum crop height [meters]", alias="h_max"
    )
    max_root_depth_max_m: float = Field(
        ..., description="Maximum root depth [meters]", alias="zr_max_max"
    )
    max_root_depth_min_m: float = Field(
        ..., description="Minimum root depth [meters]", alias="zr_max_min"
    )
    min_root_depth_m: float = Field(
        ..., description="Minimum root depth [meters]", alias="zr_min"
    )
    sowing_to_emergence_d: int = Field(
        ...,
        description="Time from sowing to emergence [days]",
        alias="l_emergence",
    )
    const_depletion_factor: float = Field(
        ..., description="Depletion factor", alias="p"
    )

    sowing_to_max_rooting_depth_d: int = Field(
        20, description="Time from sowing to maximum rooting depth [days]"
    )
    start_root_depth_pct: float = Field(
        50, description="Virtual starting root depth [%]", alias="z0_pct"
    )
    yield_max_min: float = Field(
        0.6, description="Yield max (minimum)", alias="y_max_min"
    )
    yield_max_max: float = Field(
        1.4, description="Yield max (maximum)", alias="y_max_max"
    )

    @property
    def crop_coefficient_end(self) -> float:
        return (
            self.crop_coefficient_end_min + self.crop_coefficient_end_max
        ) / 2

    @property
    def max_root_depth_m(self) -> float:
        return (self.max_root_depth_min_m + self.max_root_depth_max_m) / 2

    @property
    def yield_max(self) -> float:
        return (self.yield_max_min + self.yield_max_max) / 2



from datetime import datetime
from typing import List, cast

from datetimerange import DateTimeRange
from pydantic import BaseModel, Field


class ClimateDatum(BaseModel):
    timestamp: datetime
    u_mean_ms: float = Field(
        ..., description="Mean wind speed at 2 meters [m/s]"
    )
    RH_min: float = Field(..., description="Minimum relative humidity")
    ET_0: float = Field(..., description="Reference evapotranspiration")
    precipation_mm: float = Field(..., description="Precipitation [mm]", ge=0)


class ClimateData(object):
    def __init__(self, data: List[ClimateDatum]):
        sorted(data, key=lambda d: d.timestamp)
        self.data_points = data

    def get_data_within_range(
        self, date_range: DateTimeRange
    ) -> List[ClimateDatum]:
        """
        Get climate data within a date range
        """
        return [
            d
            for d in self.data_points
            if cast(datetime, date_range.start_datetime)
            <= d.timestamp
            <= cast(datetime, date_range.end_datetime)
        ]

    @property
    def time_range(self) -> DateTimeRange:
        """
        Get the range of the climate data
        """
        return DateTimeRange(
            self.data_points[0].timestamp, self.data_points[-1].timestamp
        )



from datetime import date, datetime
from typing import List, Optional, Tuple

from fao56_model.data_models.crop import Crop
from fao56_model.planted_crop import PlantedCrop
from fao56_model.utils import is_overlapping


class CropCalendar:
    def __init__(self, crop_plus_sowing: List[Tuple[Crop, date]]) -> None:
        self.set_crops(crop_plus_sowing)

    def set_crops(self, crop_plus_sowing: List[Tuple[Crop, date]]) -> None:
        self.planted_crops: List[PlantedCrop] = []
        for c, d in crop_plus_sowing:
            self.planted_crops.append(PlantedCrop(c, d))
        if is_overlapping([pc.growing_period for pc in self.planted_crops]):
            raise ValueError("Overlapping crop growing periods")
        sorted(self.planted_crops, key=lambda pc: pc.sowing_date)

    @property
    def time_range(self):
        period = None
        for planted_crop in self.planted_crops:
            period = (
                planted_crop.growing_period
                if period is None
                else period.encompass(planted_crop.growing_period)
            )
        return period

    def get_crop(self, timestamp: datetime) -> Optional[PlantedCrop]:
        for planted_crop in self.planted_crops:
            if timestamp in planted_crop.growing_period:
                return planted_crop
        return None



from typing import Generic, List, Optional, Type, TypeVar, Union

import pandas as pd
import pkg_resources

from fao56_model.data_models.crop import Crop, CropType
from fao56_model.data_models.soil import Soil, SoilType

T = TypeVar("T", Soil, Crop)
K = TypeVar("K", bound=Union[SoilType, CropType])


class NotFoundException(Exception):
    def __init__(self, item_type: K):
        self.item_type: K = item_type
        self.message = f"{item_type} not found in database"
        super().__init__(self.message)


class Database(Generic[T, K]):
    def __init__(self, data_type: Type[T], file_name: str):
        self.data_type: Type[T] = data_type
        self.data: List[T] = self._load_data_from_csv(file_name)

    def _load_data_from_csv(self, file_name: str) -> List[T]:
        file_path = pkg_resources.resource_filename(
            "fao56_model", f"data/{file_name}"
        )
        df = pd.read_csv(file_path)
        return [
            self.data_type(**datum)  # type: ignore
            for datum in df.to_dict(orient="records")
        ]

    def get_by_type(self, item_type: K) -> T:
        item: Optional[T] = next(
            (item for item in self.data if item.kind == item_type), None
        )
        if item is None:
            raise NotFoundException(item_type)
        return item

    def get_all(self) -> List[T]:
        return self.data


class SoilDatabase(Database[Soil, SoilType]):
    def __init__(self):
        super().__init__(Soil, "soils.csv")


class CropDatabase(Database[Crop, CropType]):
    def __init__(self):
        super().__init__(Crop, "crops.csv")
